/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.dao;

import com.portlandwebworks.commons.domain.IEntity;

import java.io.Serializable;
import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * This software is the property of Portland Webworks.
 *
 * This interface defines a standard for interfacing with a persistent data source via an ORM library.
 *
 * @Author Ronnie Northup
 */
public interface IDaoBase
{
	/**
	 * A last resort API.  This will return the underlying implementation-specific context for peforming database
	 * operations not supported by this interface.
	 *
	 * @return
	 */
	public Object getPersistenceContext();

	/**
	 * Retrieves an entity by its unique identifier.
	 *
	 * @param entityClass
	 * @param id
	 * @throws com.portlandwebworks.commons.dao.DataNotFoundException if an entity of the type requested does not exist with the
	 * 			identifier provided.
	 * @return
	 */
	public <T extends IEntity> T findById(Class<T> entityClass, Serializable id) throws DataNotFoundException;

	/**
	 * Retrieves an entity by its unique identifier, eagerly fetching those fields specified.
	 *
	 * @param entityClass
	 * @param id
	 * @param fetchOptions
	 * @throws com.portlandwebworks.commons.dao.DataNotFoundException if an entity of the type requested does not exist with the
	 * 			identifier provided.
	 * @return
	 */
	public <T extends IEntity> T findById(Class<T> entityClass, Serializable id, Enum... fetchOptions) throws DataNotFoundException;

	/**
	 *
	 * @return
	 * @param entityClass
	 */
	public <T extends IEntity> List<T> findAll(Class<T> entityClass);

	/**
	 * Returns a mapping of entities
	 *
	 * @param entityClass
	 * @param keyField
	 * @param <T>
	 * @return
	 */
	public <T extends IEntity, K extends Serializable> Map<K, T> mapAll(Class<T> entityClass, String keyField);

	/**
	 * Retrieves a collection of entities by their primary key values.
	 *
	 * @param entityClass
	 * @param ids
	 * @return
	 * @throws DataNotFoundException
	 */
	public <T extends IEntity> List<T> findAllByIds(Class<T> entityClass, Serializable[] ids) throws DataNotFoundException;

	/**
	 * Generic method to return a collection of entities that match the criteria provided.
	 * @param entityClass
	 * @param criteria
	 * @return
	 */
	public <T extends IEntity> List<T> findAllByCriteria(Class<T> entityClass, SearchCriteria criteria);

	/**
	 * Rudimentary API to retrieve a list of entities filtered using one or more predicates ANDed together.
	 *
	 * @param entityClass
	 * @param predicates
	 * @param <T>
	 * @return
	 */
	public <T extends IEntity> List<T> findAllByPredicate(Class<T> entityClass, Predicate... predicates);

	/**
	 * Rudimentary API to retrieve a single entity filtered using one or more predicates ANDed together.
	 * 
	 * @param entityClass
	 * @param predicates
	 * @param <T>
	 * @return
	 * @throws com.portlandwebworks.commons.dao.DataNotFoundException if no matching entity is found.
	 */
	public <T extends IEntity> T findByPredicate(Class<T> entityClass, Predicate... predicates) throws DataNotFoundException;

	/**
	 * Generic method to return a "page" of search results suitable for display in a grid view.
	 *
	 * @param entityClass
	 * @param criteria
	 * @param options
	 * @param fetchOptions
	 * @return
	 */
	public <T extends IEntity> GridDisplayResults<T> search(Class<T> entityClass, SearchCriteria criteria, GridDisplayOptions options, Enum... fetchOptions);


	/**
	 * Generic method to return a page of search results suitable for display in a grid view.  This specific
	 * API allows search criteria to be applied to one class, and to return as search results another class.  This
	 * allows one to run a complex set of query criteria against a domain entity, and to return as paged results
	 * a "flattened" view version of the entity.
	 * 
	 * @param viewClass         The type of each return value returned by the search method.  This is useful for
	 *                          generating "flattened" results of a query against an entity type that has a lot of
	 *                          aggregate data, or for scenarios where the rendered view requires complex calculations
	 *                          on the queried entities.  The unique ID of this class must correspond to the unique ID
	 *                          of the criteria class.
	 * 
	 * @param criteriaClass     The entity which is queried, that is, to which the search criteria will be applied.  The
	 *                          unique ID of this class must correspond to the unique ID of the view class.
	 *
	 * @param searchCriteria
	 * @param options
	 * @param fetchOptions
	 * @param <T>
	 * @return
	 */
	<T extends IEntity, Y extends IEntity> GridDisplayResults<T> search(Class<T> viewClass, Class<Y> criteriaClass, SearchCriteria searchCriteria, GridDisplayOptions options, Enum... fetchOptions);

	/**
	 * Inserts a new instance of entity into the data source.
	 *
	 * @param entity
	 * @return
	 * @throws DuplicateDataException
	 */
	public <T extends IEntity> Serializable insert(T entity) throws DuplicateDataException;

	/**
	 * Added because Jacob doesn't like writing code.
	 * 
	 * @param entity
	 * @param <T>
	 * @return
	 * @throws DuplicateDataException
	 */
	public <T extends IEntity> void saveOrUpdate(T entity) throws DuplicateDataException;

	/**
	 * Propagates local entity changes to the underlying database.  The actual propagation may not happen until the end
	 * of the transaction, depending upon how the ORM implementation behaves.
	 *
	 * @param entity
	 * @throws DuplicateDataException
	 */
	public <T extends IEntity> void update(T entity) throws DuplicateDataException;

	/**
	 * Removes an entity from the data source.
	 *
	 * @param entity
	 */
	public <T extends IEntity> void delete(T entity);

	/**
	 * Removes an entity from the data source that has not been loaded by referencing it by its identifier.
	 * @param entityClass
	 * @param id
	 */
	public <T extends IEntity> void deleteById(Class<T> entityClass, Serializable id);

	/**
	 * Removes a collection of entities from the data source.
	 * @param entities
	 */
	public <T extends IEntity> void deleteAll(Collection<T> entities);

	/**
	 * A generic way of retrieving a sequential value from the database that is guaranteed to be unique within the
	 * scope of the counter name provided.
	 *
	 * @param counterName
	 * @return
	 * @deprecated  This method should be phased out in favor of calling #executeSQLQuery
	 */
	@Deprecated
	public Long getCounterValue(String counterName);

	/**
	 * This should force the DAO implementation to flush all pending changes to the database without yet committing.
	 */
	public void saveChanges();

	/**
	 * Executes a text query in the underlying persistence query language, returning a list of entities of the specified
	 * type.
	 *
	 * @param query         The query text used to execute the query, using the language of the configured persistence
	 *                      implementation.
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 *
	 * @return
	 */
	public <T extends IEntity> List<T> executeQuery(String query, Object... params);

	/**
	 * Executes a query that updates records in the database.
	 *
	 * @param hql           The implementation-specific query to execute.
	 *
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return              The total number of objects affected by the update.
	 */
	public int executeUpdate(String hql, Object... params);

	/**
	 *
	 * @param query
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return
	 */
	public List<Map<String, ?>> executeQueryProjections(String query, Object... params);

	/**
	 * 
	 * @param query
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return
	 */
	public <T> List<T> executeQueryProjection(String query, Object... params);

	/**
	 * Executes a "named" query using the underlying persistence's query configuration mechanism, returning a list of
	 * entities of the specified type.
	 *
	 * @param name          The configured name of the query.
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return
	 */
	public <T extends IEntity> List<T> executeNamedQuery(String name, Object params);

	/**
	 * 
	 * @param query
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return
	 */
	public List<Map<String, ?>> executeNamedQueryProjections(String query, Object params);

	/**
	 * Executes an textual query in the underlying persistence query language, returning a list of singularly projected
	 * results of the specified type.
	 *
	 * @param query         The query text used to execute the query, using the language of the configured persistence
	 *                      implementation.
	 * @param returnType    The expected type of the return values.
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return
	 */
	public <T> List<T> executeNamedQueryProjection(String query, Class returnType, Object... params);

	/**
	 * Executes a raw SQL query, returning a list of Maps, each a set of name value pairs representing one projected
	 * result.  Highly recommended that you adhere to the ANSI-92 SQL standard to allow your code to work with different
	 * persistence implementations.
	 *
	 * Consider using named SQL queries instead of inline queries.
	 *
	 * @param sql           The SQL statement to be executed, with any named parameters formatted as :name
	 * @param returnValues  The map of return values expected in the query result, where the key in the map is the name
	 *                      of the selected field or alias, and the value of the map is the Class of the expected value.
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return
	 */
	public List<Map<String, ?>> executeSQL(String sql, Map<String, Class<?>> returnValues, Object... params);

	/**
	 * Executes a raw SQL query, returning a list of singular values.  Highly recommended that you adhere
	 * to the ANSI-92 SQL standard to allow your code to work with different persistence implementations.
	 *
	 * Consider using named SQL queries instead of inline queries.
	 *
	 * @param sql           The SQL statement to be executed, with any named parameters formatted as :name
	 * @param returnAlias
	 * @param returnType    The expected type of the return values.
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 */
	public <T> List<T> executeSQLQuery(String sql, String returnAlias, Class returnType, Object... params);

	/**
	 * Executes an SQL statement that updates or deletes records in the database.
	 *
	 * @param sql           The SQL, preferably ANSI-92 compliant
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return              The total number of records affected by the SQL statement execution.
	 */
	public int executeSQLUpdate(String sql, Object... params);

	/**
	 * Executes a raw SQL query, returning a list of entities of the specific type.  Highly recommended that you adhere
	 * to the ANSI-92 SQL standard to allow your code to work with different persistence implementations.
	 *
	 * If the SQL query has 
	 *
	 * @param entityClass
	 * @param sql           The SQL query to execute, with parameters formatted as :value
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return
	 */
	public <T extends IEntity> List<T> executeSQLQuery(Class<T> entityClass, String sql, Object... params);

	/**
	 * Executes a raw SQL query stored in a configuration file that has been loaded by the persistence engine.
	 *
	 * @param name
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return
	 */
	public List<Map<String, ?>> executeNamedSQL(String name, Object... params);

	/**
	 * 
	 *
	 * @param name
	 * @param params        Any query parameters needed to execute the query.  Query parameters can consist of any of the
	 *                      following types of data:
	 *                      <ul>
	 *                          <li>An object that exposes java bean properties that can be used to supply values for the
	 *                              named query parameters in the query provided.</li>
	 *                          <li>A Map implementation whose key/value pairs can be used to supply values for the named
	 *                              query parameters in the query provided.</li>
	 *                          <li>A list of vararg values that correspond in order to the ? style query parameters in
	 *                              the query provided.</li>
	 *                      </ul>
	 * @return
	 */
	public <T extends IEntity> List<T> executeNamedEntitySQL(String name, Object... params);
}
