/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
package com.portlandwebworks.commons.hibernate;

import com.portlandwebworks.commons.dao.*;
import com.portlandwebworks.commons.domain.IEntity;
import com.portlandwebworks.commons.reflection.FieldUtils;
import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.hibernate.*;
import org.hibernate.criterion.*;
import org.hibernate.engine.NamedSQLQueryDefinition;
import org.hibernate.engine.query.sql.NativeSQLQueryReturn;
import org.hibernate.engine.query.sql.NativeSQLQueryScalarReturn;
import org.hibernate.exception.ConstraintViolationException;
import org.hibernate.impl.SessionFactoryImpl;
import org.hibernate.type.Type;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.orm.hibernate3.SessionFactoryUtils;

import javax.persistence.Column;
import javax.persistence.EmbeddedId;
import java.io.Serializable;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.math.BigDecimal;
import java.text.MessageFormat;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * @Author rnorthup
 */
public class HibernateDaoBase implements IDaoBase {
	private Logger logger = Logger.getLogger(this.getClass());

	private SessionFactory sessionFactory;

	private HashMap<Class<?>, List<String>> fieldMultiplicities = new HashMap<Class<?>, List<String>>();

	protected String entityName;

	private static Map<Class<?>, Type> typeMappings = new HashMap<Class<?>, Type>();

	static {

		// Generate a lookup table of all Hibernate types keyed by java type,  Why doesn't the Hibernate API have this?
		for (Field field : Hibernate.class.getFields()) {
			field.setAccessible(true);
			if (Modifier.isStatic(field.getModifiers())) {
				if (Type.class.isAssignableFrom(field.getType())) {
					try {
						Type type = (Type) field.get(null);
						typeMappings.put(type.getReturnedClass(), type);
					}
					catch (IllegalAccessException e) {
					}
				}
			}
		}
	}

	/**
	 * Determines which Hibernate type to use based on the java type provided
	 *
	 * @param clazz
	 * @return
	 */
	protected static Type getTypeForClass(Class<?> clazz) {
		Type type = typeMappings.get(clazz);
		if (type != null) {
			return type;
		}
		return Hibernate.STRING;
	}

	/**
	 *
	 */
	public HibernateDaoBase() {
	}

	/**
	 * This session factory is used by the DAO for all database access and should be set by the container at runtime.
	 *
	 * @param sessionFactory
	 */
	@Autowired(required = true)
	public void setSessionFactory(SessionFactory sessionFactory) {
		this.sessionFactory = sessionFactory;
	}

	/**
	 * The HQL name of the entity.
	 *
	 * @return
	 */
	public String getEntityName(IEntity entityInstance) {
		if (entityName == null) {
			entityName = getSession().getEntityName(entityInstance);
		}
		return entityName;
	}

	/**
	 * @param entityClass
	 * @return
	 */
	protected boolean hasMultiValueField(Class entityClass) {
		if (!fieldMultiplicities.containsKey(entityClass)) {
			//process this class
			List<String> fieldNames = DaoUtils.getMultiValueFieldNames(entityClass);
			fieldMultiplicities.put(entityClass, fieldNames);
		}
		return fieldMultiplicities.get(entityClass).size() > 0;
	}

	/**
	 * Determines if the property represented by the enumeration is backed by a Collection.
	 *
	 * @param entityClass
	 * @param fieldName
	 * @return
	 */
	protected <T extends IEntity> boolean isMultiValueField(Class<T> entityClass, String fieldName) {
		if (!fieldMultiplicities.containsKey(entityClass)) {
			//process this class
			List<String> fieldNames = DaoUtils.getMultiValueFieldNames(entityClass);
			fieldMultiplicities.put(entityClass, fieldNames);
		}

		return fieldMultiplicities.get(entityClass).contains(fieldName);
	}

	/**
	 * Utility method to retrieve the default schema configured for the underlying Hibernate session.
	 *
	 * @return The string name of the schema, or <code>null</code> if no default schema is configured.
	 */
	protected String getDefaultSchemaName() {
		// This interface doesn't have the necessary api.  Need to downcast to the impl.
		SessionFactory factory = getSession().getSessionFactory();
		if (factory instanceof SessionFactoryImpl) {
			SessionFactoryImpl impl = (SessionFactoryImpl) factory;
			return impl.getSettings().getDefaultSchemaName();
		}
		return null;
	}

	/**
	 * Provides DAO subclasses with the ability to access the underlying hibernate session.
	 *
	 * @return
	 */
	protected Session getSession() {
		return SessionFactoryUtils.getSession(sessionFactory, true);
	}

	/**
	 * Exposes the Hibernate session to the user of the API as a last resort to accomplish tasks not supported by the
	 * interface.
	 *
	 * @return
	 */
	public Object getPersistenceContext() {
		return getSession();
	}

	/**
	 * Helper method for reducing the list results of a query down to a single result.
	 *
	 * @param query
	 * @param <T>
	 * @return
	 */
	@SuppressWarnings("unchecked")
	protected <T> T getSingleResult(Criteria query) throws DataNotFoundException, DataIntegrityException {
		List result = query.list();
		switch (result.size()) {
			case 0:
				throw new DataNotFoundException();
			case 1:
				return (T) result.get(0);
			default:
				throw new DataIntegrityException("More than one result found for query (expected one)");
		}
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> T findById(Class<T> entityClass, Serializable id) throws DataNotFoundException, DataIntegrityException {
		T result = (T) getSession().get(entityClass, id);
		if (result == null) {
			throw new DataNotFoundException(entityClass, id);
		}
		return result;
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> T findById(Class<T> entityClass, Serializable id, Enum... fetchOptions) throws DataNotFoundException {
		Criteria criteria = getSession().createCriteria(entityClass);

		Enum prefetch = applyPrefetch(entityClass, criteria, fetchOptions);

		criteria.add(Restrictions.eq("id", id));
		criteria.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		T entity = (T) criteria.uniqueResult();

		if (entity == null) {
			throw new DataNotFoundException(entityClass, id);
		}

		applyPostfetch(entityClass, prefetch, entity, fetchOptions);

		return entity;
	}

	@SuppressWarnings({"unchecked"})
	public <T extends IEntity> List<T> findAll(Class<T> entityClass) {
		Criteria query = getSession().createCriteria(entityClass);

		applySorting(entityClass, query, null);

		return query.list();
	}

	public <T extends IEntity, K extends Serializable> Map<K, T> mapAll(Class<T> entityClass, String keyField) {
		Criteria query = getSession().createCriteria(entityClass);

		applySorting(entityClass, query, null);

		List<T> results = query.list();
		Map<K, T> resultMap = new HashMap<K, T>();
		try {
			for (T result : results) {
				resultMap.put(
						(K) PropertyUtils.getProperty(result, keyField),
						result
				);
			}
		}
		catch (Exception e) {
			throw new RuntimeException("Ack!", e);
		}
		return resultMap;
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> List<T> findAllByIds(Class<T> entityClass, Serializable[] ids) throws DataNotFoundException {
		Criteria query = getSession().createCriteria(entityClass);

		query.add(Restrictions.in("id", ids));

		query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		List list = query.list();

		if (list.size() < ids.length) {
			// Returned entity list size is less than size of requested ID array.  This means that at least one ID does
			// not represent a persisted entity.  Report an error by identifying the first ID that doesn't match the
			// ID of any returned entity.
			for (Serializable id : ids) {
				boolean match = false;
				for (Object obj : list) {
					IEntity entity = (IEntity) obj;
					if (entity.getId().equals(id)) {
						match = true;
						break;
					}
				}
				if (!match) {
					throw new DataNotFoundException(entityClass, id);
				}
			}
		}
		return list;
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> List<T> findAllByCriteria(Class<T> entityClass, SearchCriteria searchCriteria) {
		Criteria query = getSession().createCriteria(entityClass);

		applySearchCriteria(searchCriteria, query);

		applySorting(entityClass, query, null);

		// Applying predicates may cause joins to occur if a predicate spans entity relationships.
		query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		return query.list();
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> List<T> findAllByPredicate(Class<T> entityClass, Predicate... predicates) {
		Criteria query = getSession().createCriteria(entityClass);

		applyPredicates(query, Arrays.asList(predicates));

		// Applying predicates may cause joins to occur if a predicate spans entity relationships.
		query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		applySorting(entityClass, query, null);

		return query.list();
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> T findByPredicate(Class<T> entityClass, Predicate... predicates) throws DataNotFoundException {
		Criteria query = getSession().createCriteria(entityClass);

		applyPredicates(query, Arrays.asList(predicates));

		// Applying predicates may cause joins to occur if a predicate spans entity relationships.
		query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		List<T> results = query.list();
		switch (results.size()) {
			case 0:
				return null;
			case 1:
				return results.get(0);
			default:
				throw new DataIntegrityException("Multiple values found query, expected zero or one");
		}
	}

	protected <T extends IEntity> Enum applyPrefetch(Class<T> entityClass, Criteria criteria, Enum... fetchOptions) {
		// This will be the singular 1-M entity field that is pre-fetched.  Hibernate only allows
		// pre-fetching one aggregate field when retrieving an entity.
		Enum prefetch = null;

		// Iterate through the fetch options, eager joining all 1-1 relationships, an skipping over any
		// 1-M relationships after the first one has been eager joined.
		for (Enum fetchOption : fetchOptions) {

			// The name of the relationship to fetch is not the same as the name of the annotation.
			FetchOption fetch = fetchOption.getClass().getAnnotation(FetchOption.class);
			if (fetch != null) {
				String join = fetch.value();
				logger.debug("Eager joining relationship: " + join);
				criteria.setFetchMode(join, FetchMode.JOIN);
			} else {
				if (isMultiValueField(entityClass, fetchOption.toString()) && prefetch != null) {
					logger.debug("Skipping fetch option " + fetchOption + " as eager load field in query: can only eager load one 1-M field (" + prefetch + ")");
				} else {
					logger.debug("Eager joining collection: " + fetchOption);
					criteria.setFetchMode(fetchOption.toString(), FetchMode.JOIN);
					prefetch = fetchOption;
				}
			}
		}

		return prefetch;
	}

	/**
	 * Eager loads all multi-valued fields on the entity provided that are represented by the fetch options specified,
	 * excluding the field that was previously fetched by join.
	 *
	 * @param entityClass
	 * @param prefetch
	 * @param entity
	 * @param fetchOptions
	 * @param <T>
	 */
	protected <T extends IEntity> void applyPostfetch(Class<T> entityClass, Enum prefetch, T entity, Enum... fetchOptions) {
		String field = null;

		try {
			// Post-fetch all requested aggregate fields except for the one that was pre-fetched.  The name
			// of the enum is the same as the bean property.
			for (Enum fetchOption : fetchOptions) {
				if (isMultiValueField(entityClass, fetchOption.toString()) && fetchOption != prefetch) {
					field = fetchOption.toString();
					logger.debug("Eager loading collection after query: " + field);
					Object obj = PropertyUtils.getProperty(entity, field);
					if (obj instanceof Collection)
						((Collection) obj).size();
					else if (obj instanceof Map)
						((Map) obj).size();
				}
			}
		}
		catch (Exception e) {
			throw new RuntimeException("Unhandled exception fetching field " + field, e);
		}

	}

	protected void applyPredicates(Criteria query, List<Predicate> predicates) {
		// This is the set of all entities that have been joined in the search query, mapped to their aliases.
		HashMap<String, String> joinedEntities = new HashMap<String, String>();

		// This guarantees a unique alias name throught the building of the query.
		int aliasCount = 0;

		applyPredicates(query, null, predicates, joinedEntities, aliasCount);
	}

	/**
	 * Internal convenience method for applying predicates to a hibernate Criteria instance.
	 *
	 * @param query		  The criteria to which the predicates will be applied
	 * @param junction	   The current junction that is being assebled, if any.  If a junction is specified, the
	 *                       predicates will be applied to it, not the criteria object.
	 * @param predicates
	 * @param joinedEntities
	 * @param aliasCount
	 * @return The current number of aliases that have been created.
	 */
	private int applyPredicates(Criteria query, Junction junction, List<Predicate> predicates, HashMap<String, String> joinedEntities, int aliasCount) {
		for (Predicate predicate : predicates) {

			// Check the predicate for nested predicates. If any are found, ignore the basic properties for this predicate
			// and treat it as an AND or OR operation on a set of nested predicates.
			Predicate[] nestedPredicates = predicate.getPredicates();
			if (!ArrayUtils.isEmpty(nestedPredicates)) {

				// Are the nested predicates to be ANDed or ORed in the query?
				boolean isAnd = predicate.getAnd();

				// Recursively call this method, passing on an AND or OR junction to capture the nested predicates.
				Junction subJunction = isAnd ? Restrictions.conjunction() : Restrictions.disjunction();
				aliasCount = applyPredicates(query, subJunction, Arrays.asList(nestedPredicates), joinedEntities, aliasCount);

				// Once the nested predicates are handled, apply them to the query or junction currently being worked on.
				if (junction != null) {
					junction.add(subJunction);
				} else {
					query.add(subJunction);
				}
			} else {
				// This predicate has no nested predicates.  It is a simple property-operator-operand predicate.
				String property = predicate.getProperty();
				String operand = property;
				String joinPath = predicate.getJoinPath();

				// Check for dot notation on the property field.  This is acceptable in terms of API usage because we
				// can easily move the relationship defined on the property to be the join path.  In this way we can
				// phase out the need to have the predicate separate out the join path from the target property.
				int dotpos = StringUtils.lastIndexOf(property, '.');
				if (dotpos > -1) {
					joinPath = StringUtils.defaultString("") + property.substring(0, dotpos);
					property = property.substring(dotpos + 1);
				}

				if (joinPath != null) {

					// This will be the alias for the last joined entity, to be used in the next join statement.
					String lastAlias = null;

					// Break the join path into its individual entity relationship traversals.
					final String[] pathElements = joinPath.split("\\.");

					// This will represent the full joined path, so far.
					StringBuilder pathLookupKey = new StringBuilder();

					// Iterate through each entity relationship, ensuring that it has been joined by a previous predicate
					// and if not, do so now.
					for (String subPath : pathElements) {

						// Assemble the join path so far, using dot notation between joins
						if (pathLookupKey.length() > 0) {
							pathLookupKey.append(".");
						}

						pathLookupKey.append(subPath);

						// Check the lookup table to see if this path has already been joined and the entity aliased.
						final String lookupKey = pathLookupKey.toString();
						String alias = joinedEntities.get(lookupKey);
						if (alias == null) {

							// Build an entity name that is guaranteed to be unique using an alias counter.
							alias = StringUtils.substring(subPath, 0, 3) + "_" + aliasCount++;

							if (lastAlias != null) {
								// Join using the previously aliased entity as the reference point for the join
								query.createAlias(lastAlias + "." + subPath, alias);
							} else {
								// Join using the implicit source entity as the reference point for the join.
								query.createAlias(subPath, alias);
							}

							joinedEntities.put(lookupKey, alias);
						}

						lastAlias = alias;
					}

					// Finally, the operand for this predicate is the last generated entity plus target property.
					operand = MessageFormat.format("{0}.{1}", lastAlias, property);
				}

				// Convert the predicate into a Hibernate criterion.  if we are currently working on a junction, add it
				// to the junction, otherwise add it to the base criteria object.
				final Criterion criterion = getCriterion(operand, predicate.getOperator(), predicate.getValue());
				if (junction != null) {
					junction.add(criterion);
				} else {
					query.add(criterion);
				}
			}
		}
		return aliasCount;
	}

	/**
	 * Convenience method for applying query search criteria to a query by scanning an object for @SearchCriterion
	 * annotated fields with non-empty values, and converting them into Criteria API calls.
	 *
	 * @param searchCriteria
	 * @param query
	 */
	protected void applySearchCriteria(SearchCriteria searchCriteria, Criteria query) {
		if (searchCriteria == null) {
			throw new IllegalArgumentException("Search criteria is required");
		}

		if (logger.isDebugEnabled()) {
			//TODO:
		}

		List<Predicate> predicates = null;
		Field currField = null;

		try {
			// Iterate through each field on the search criteria object and check it for any of the annotations that
			// support search criteria.
			for (Field field : FieldUtils.getDeclaredFields(searchCriteria.getClass())) {

				// For error reporting
				currField = field;

				// Check to see if there's a single SearchCriteria annotation on this field.  If there is, build the
				// equivalent predicate for it and add it to the list of predicates.
				SearchCriterion criterion = field.getAnnotation(SearchCriterion.class);
				if (criterion != null) {

					assertFieldIsNonPrimitive(field);

					Object value = field.get(searchCriteria);

					Predicate predicate = buildPredicate(field, criterion, value);
					if (predicate != null) {
						// Lazy initialization of the predicates list to handle the situation where the majority of searches
						// performed have no search criteria.
						if (predicates == null) {
							predicates = new ArrayList<Predicate>();
						}
						predicates.add(predicate);
					}
				} else {
					SearchCriterion[] subCriteria = null;
					boolean and = true;

					// If this annotation appears on a field, all of its nested SearchCriterion annotations will be ANDed
					// together in one predicate.
					AllSearchCriteria andOp = field.getAnnotation(AllSearchCriteria.class);
					if (andOp != null) {
						subCriteria = andOp.value();
						and = true;
					}

					// If this annotation appears on a field, all of its nested SearchCriterion annotations will be ORed
					// together in one predicate.
					AnySearchCriteria orOp = field.getAnnotation(AnySearchCriteria.class);
					if (orOp != null) {
						subCriteria = orOp.value();
						and = false;
					}

					// If an "all" or "any" was found, assemble the nested predicates under one predicate.
					if (subCriteria != null) {
						List<Predicate> subPreds = null;

						for (SearchCriterion searchCriterion : subCriteria) {

							assertFieldIsNonPrimitive(field);

							Object value = field.get(searchCriterion);

							// Attempt to build a predicate for the field, and if one is returned add it to the sublist
							Predicate pred = buildPredicate(field, criterion, value);
							if (pred != null) {
								// Lazy initialization of list for nested predicates.
								if (subPreds == null) {
									subPreds = new ArrayList<Predicate>();
								}
								subPreds.add(pred);
							}
						}

						if (!CollectionUtils.isEmpty(subPreds)) {
							// build the appropriate predicate for doing AND or OR operation.
							Predicate multiPredicate = and ? Predicate.and(subPreds) : Predicate.or(subPreds);

							if (predicates == null) {
								predicates = new ArrayList<Predicate>();
							}
							predicates.add(multiPredicate);
						}
					}
				}
			}
		}
		catch (IllegalAccessException e) {
			throw new RuntimeException(String.format("Illegal access to %s.%s", searchCriteria.getClass().getName(), currField.getName()));
		}

		// If this list is non-null, at least one predicate was generated from search criteria.
		if (predicates != null) {
			applyPredicates(query, predicates);
		}
	}

	/**
	 * Throws an exception if a field represents a primitive value.
	 *
	 * @param field The field to evaluate.
	 * @throws RuntimeException if the field has a primitive value.
	 */
	private void assertFieldIsNonPrimitive(Field field) {
		if (field.getType().isPrimitive()) {
			throw new RuntimeException("Primitive types cannot be used as search criteria.  Impossible to tell if a primitive type is excluded from a search.");
		}
	}

	/**
	 * Builds a Predicate instance from a SearchCriterion annotation.  The primary purpose of this method is to transform
	 * search criteria functionality into predicate functionality for plugging into the findByPredicate() APIs.
	 *
	 * @param field	 The field from which the annotation will be built.
	 * @param criterion The annotation for which the predicate will be built.
	 * @param value	 The value assigned to the field provided.
	 * @return A new instance of Predicate mirroring the properties of the annotation, or <code>null</code> if
	 *         the field's value is null or is an empty string and the annotation was configured to ignore
	 *         empty strings.
	 */
	protected Predicate buildPredicate(Field field, SearchCriterion criterion, Object value) {
		// Check to see if the value is null, and if so don't include it in the query.
		if (value == null) {
			return null;
		}

		// Don't apply an empty string if the annotation says not to do so.
		if (value instanceof String && !criterion.useEmptyStringValue() && value.equals("")) {
			return null;
		}

		// The operand is in the propertyName field.  If no property name specified in annotation, default to field name.
		String operand = criterion.propertyName();
		if (operand.equals("")) {
			operand = field.getName();
		}

		if (StringUtils.isNotBlank(criterion.joinPath())) {
			// Build a predicate with a join path in it.
			return new Predicate(criterion.joinPath(), operand, criterion.operator(), value);
		} else {
			// Build a simply property predicate with no join path.
			return new Predicate(operand, criterion.operator(), value);
		}
	}


	/**
	 * Unified means of applying query parameters to a query.
	 *
	 * @param query  The query whose placeholders will be
	 * @param params Either a Map of name value pairs, or an object whose public fields, both declared and inherited,
	 *               that will be accessed as needed by the placeholders in the query provided.
	 */
	protected void applyParamsToQuery(Query query, Object... params) {
		if (params.length > 0) {

			// If the query contains no named parameters, assume that the query contains ? placeholders for varargs values
			// Apply each value in order of position in the array.
			if (ArrayUtils.isEmpty(query.getNamedParameters())) {
				int index = 0;
				for (Object param : params) {
					query.setParameter(index++, param);
				}
			} else {
				// Allow the caller to send a list of map entries to the varargs parameter.
				if (Map.Entry.class.isAssignableFrom(params[0].getClass())) {
					for (Object param : params) {
						Map.Entry entry = (Map.Entry) param;
						query.setParameter(entry.getKey().toString(), entry.getValue());
					}
				} else if (Map.class.isAssignableFrom(params[0].getClass())) {
					// A Map was passed in.
					query.setProperties((Map) params[0]);
				} else {
					// A bean was passed to the varargs.
					query.setProperties(params[0]);
				}
			}
		}
	}

	/**
	 * Converts a predicate object into the appropriate hibernate Criteria expression.
	 *
	 * @param operand
	 * @param operator
	 * @param value
	 * @return
	 */
	protected Criterion getCriterion(String operand, Operator operator, Object value) {
		switch (operator) {
			// Convert the generic DAO operation into a hibernate restriction and add it to the
			// query.
			case EqualTo:
				return Restrictions.eq(operand, value);
			case GreaterThan:
				return Restrictions.gt(operand, value);
			case GreaterThanOrEqual:
				return Restrictions.ge(operand, value);
			case NotEqualTo:
				return Restrictions.ne(operand, value);
			case IsNotNull:
				return Restrictions.isNotNull(operand);
			case IsNull:
				return Restrictions.isNull(operand);
			case IsIn:
				Class operandType = value.getClass();
				if (Collection.class.isAssignableFrom(operandType)) {
					return Restrictions.in(operand, (Collection) value);
				} else if (operandType.isArray()) {
					Class arrayType = operandType.getComponentType();
					if (arrayType.isPrimitive()) {
						// primitive array needs to be converted into object array
						int length = Array.getLength(operand);
						Object[] arrayValues = new Object[length];
						for (int i = 0; i < Array.getLength(operand); i++) {
							arrayValues[i] = Array.get(operand, i);
						}
						return Restrictions.in(operand, arrayValues);
					} else {
						return Restrictions.in(operand, (Object[]) value);
					}
				}
			case LessThan:
				return Restrictions.lt(operand, value);
			case LessThanOrEqual:
				return Restrictions.le(operand, value);
			case Matches:
				return Restrictions.ilike(operand, value.toString(), MatchMode.ANYWHERE);
			case StartsWith:
				return Restrictions.ilike(operand, value.toString(), MatchMode.START);
			case EndsWith:
				return Restrictions.ilike(operand, value.toString(), MatchMode.END);
			case IsLike:
				return Restrictions.ilike(operand, value.toString().replace('*', '%'));
			case CountEqualTo:
				return Restrictions.sizeEq(operand, ((Number) value).intValue());
			case CountGreaterThan:
				return Restrictions.sizeGt(operand, ((Number) value).intValue());
			case CountLessThan:
				return Restrictions.sizeLt(operand, ((Number) value).intValue());
			case CountNotEqualTo:
				return Restrictions.sizeNe(operand, ((Number) value).intValue());
			case CountGreaterThanOrEqual:
				return Restrictions.sizeGe(operand, ((Number) value).intValue());
			case CountLessThanOrEqual:
				return Restrictions.sizeLe(operand, ((Number) value).intValue());
			default:
				return null;
		}
	}

	public <T extends IEntity> GridDisplayResults<T> search(Class<T> entityClass, SearchCriteria searchCriteria, GridDisplayOptions displayOptions, Enum... fetchOptions) {
		return search(entityClass, entityClass, searchCriteria, displayOptions, fetchOptions);
	}

	public <T extends IEntity, Y extends IEntity> GridDisplayResults<T> search(Class<T> viewClass, Class<Y> criteriaClass, SearchCriteria searchCriteria, GridDisplayOptions displayOptions, Enum... fetchOptions) {
		GridDisplayResults<T> results = new GridDisplayResults<T>();

		Enum prefetch = null;

		// If display options indicate that the query results should be a page of partial results, run this query in
		// two phases.  First phase is to execute a count of distinct IDs using the same search criteria as the results
		// query.  Second phase is to perform the results query, with paging applied.  This allows the method to return
		// not only the page of results but the total number of matching records.
		if (displayOptions != null && displayOptions.getPageSize() != GridDisplayOptions.ALL_RESULTS) {

			Criteria countQuery = getSession().createCriteria(criteriaClass);

			applySearchCriteria(searchCriteria, countQuery);

			// Return a count, not the actual entity list.
			Projection countProjection;


			Field embedded = DaoUtils.getEmbeddedIdField(criteriaClass);

			if (embedded != null) {
				countProjection = createCompositeIdCountProjection(embedded);
			} else {
				countProjection = Projections.countDistinct("id");
			}

			countQuery.setProjection(countProjection);

			// Save this value on the return object.
			Number count = (Number) countQuery.uniqueResult();
			results.setTotalItemCount(count.intValue());

			// Now run the actual query, applying paging parameters.


			boolean entityHasEmbeddedId = DaoUtils.getEmbeddedIdField(viewClass) == null ? false : true;


			//TODO: find a way to unify these two query approaches into one.  It shouldn't be this difficult to come up
			//TODO: with a distinct set of results for a hibernate query.  Perhaps add all entity fields to a projection
			//TODO: list with a distinct projection wrapper around it?
			if (entityHasEmbeddedId || (!hasMultiValueField(viewClass) && viewClass.equals(criteriaClass))) {
				// This is the less than ideal way to run this query, necessary because the "ideal" way doesn't work
				// with a composite primary key.
				Criteria resultsQuery = getSession().createCriteria(viewClass);

				prefetch = applyPrefetch(viewClass, resultsQuery, fetchOptions);

				applySearchCriteria(searchCriteria, resultsQuery);

				applySorting(viewClass, resultsQuery, displayOptions);

				applyPaging(resultsQuery, displayOptions);

				// DON'T USE THIS API--THE LIST TRIMMING FOR DISTINCT AFTER DB CALL, THROWING OFF PAGING.
				//resultsQuery.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

				results.setItems(resultsQuery.list());
			} else {
				// This is the ideal way to run this query, which assumes a single value primary key.
				DetachedCriteria resultsQuery = DetachedCriteria.forClass(criteriaClass);
				Criteria queryAdapter = new DetachedCriteriaAdapter(resultsQuery);
				applySearchCriteria(searchCriteria, queryAdapter);

				resultsQuery.setProjection(
						Projections.distinct(Projections.property("id"))
				);

				Criteria rootQuery = getSession().createCriteria(viewClass);
				prefetch = applyPrefetch(viewClass, rootQuery, fetchOptions);
				rootQuery.add(Subqueries.propertyIn("id", resultsQuery));
				applyPaging(rootQuery, displayOptions);
				applySorting(viewClass, rootQuery, displayOptions);

				results.setItems(rootQuery.list());
			}

			logger.debug("Total items matching criteria: " + results.getTotalItemCount());
		} else {
			// The simple scenario--no paging.  Returning all results.  No count query necessary.  Apply the search
			// criteria and the sorting and get all results.
			Criteria resultsQuery = getSession().createCriteria(viewClass);
			prefetch = applyPrefetch(viewClass, resultsQuery, fetchOptions);
			applySearchCriteria(searchCriteria, resultsQuery);
			applySorting(viewClass, resultsQuery, displayOptions);

			results.setItems(resultsQuery.list());

			results.setTotalItemCount(results.getItems().size());
		}

		if (fetchOptions.length > 1) {
			for (T item : results.getItems()) {
				applyPostfetch(viewClass, prefetch, item, fetchOptions);
			}
		}

		return results;
	}

	/**
	 * Thank you John Gordon for implementing proper record counting for legacy databases with composite primary keys.
	 *
	 * @param embedded The composite primary key field
	 * @return A Hibernate projection that can be applied to a Criteria instance.
	 */
	protected Projection createCompositeIdCountProjection(Field embedded) {

		// get a list of all declared fields on the composite key.
		Field[] compositeFields = FieldUtils.getDeclaredFields(embedded.getType());

		if (compositeFields.length == 0) {
			throw new IllegalArgumentException("Embedded id class contains no fields.");
		}

		String countSql = "COUNT(DISTINCT %s) AS %s";

		// This will hold the comma-delimited set of composite key names to plug into the query.
		StringBuilder fieldSql = new StringBuilder();

		for (Field compositeField : compositeFields) {

			Column columnAnnotation = compositeField.getAnnotation(Column.class);

			if (columnAnnotation != null) {

				if (fieldSql.length() > 0) {
					fieldSql.append(", ");
				}
				// This field is part of the primary key.  Add it to the list.
				String columnName = columnAnnotation.name();
				if (StringUtils.isNotEmpty(columnName)) {
					fieldSql.append(columnName);
				} else {
					fieldSql.append(compositeField.getName());
				}
			} else {
				// R.N. 1/29/2011 - I disabled this line of JG code becasue I don't think you should include a field in
				// the list of composite key names if the field has no @Column annotation on it.
				//fieldSql.append(compositeField.getName());
				logger.warn(String.format("Skipping over field %s in composite key class %s--it has no @Column annotation ",
						compositeField.getName(), embedded.getName()));
			}
		}

		String[] rowAliases = new String[]{"resultCount"};

		countSql = String.format(countSql, fieldSql.toString(), rowAliases[0]);

		Type[] types = new Type[]{HibernateDaoBase.getTypeForClass(Integer.class)};

		return Projections.sqlProjection(countSql, rowAliases, types);

	}

	protected void applyPaging(Criteria resultsQuery, GridDisplayOptions options) {
		if (logger.isDebugEnabled() && options != null) {
			logger.debug(String.format("Applying paging: index=%d, size=%d", options.getPageIndex(), options.getPageSize()));
		}

		// Skip ahead to page requested.
		resultsQuery.setFirstResult(options.getPageIndex() * options.getPageSize());

		resultsQuery.setMaxResults(options.getPageSize());
	}

	/**
	 * Convenience method for applying order by options to a query.
	 *
	 * @param criteria
	 * @param options
	 */
	protected <T extends IEntity> void applySorting(Class<T> entityClass, Criteria criteria, GridDisplayOptions options) {
		// Apply sorting by first checking for a parameter value, if not present use a default sort value
		// defined on the entity, otherwise use the id field.
		String sortProperty = null;
		Boolean sortAscending = null;

		// If the options parameter is non-null, attempt to get some sort stuff from it.
		if (options != null) {
			sortProperty = options.getSortProperty();
			sortAscending = options.isAscending();
		}

		if (sortProperty == null) {
			applyDefaultSortToEntity(entityClass, criteria);
			return;
		}

		String[] sortProperties = StringUtils.deleteWhitespace(sortProperty).split(",");

		List<String> tableAliases = new ArrayList<String>();

		for (String currentProperty : sortProperties) {

			String compoundKeySortProperty = getSortStringForCompoundKey(entityClass, currentProperty);

			//NB:  compound key and joined sortProperty string both have a "." in it.
			//don't call the second sort call if the 1st returned something.  The sort will
			//either be an embedded or a joined, but not both.
			if (compoundKeySortProperty == null) {

				currentProperty = getSortPropertyForJoinedFieldAndCreateTableAlias(criteria,
						tableAliases, currentProperty);
				
			}


			if (logger.isDebugEnabled() && options != null) {
				logger.debug(MessageFormat.format("Applying sorting: property={0}, ascending={1}",
						currentProperty, sortAscending));
			}

			if (sortAscending == null) {
				sortAscending = DaoUtils.getDefaultSortAscending(entityClass);
			}

			if (sortAscending) {
				criteria.addOrder(Order.asc(currentProperty));
			} else {
				criteria.addOrder(Order.desc(currentProperty));
			}
		}
	}

	private <T extends IEntity> String getSortStringForCompoundKey(Class<T> entityClass, String sortProperty) {
		//TODO: this might be an expensive operation for each search.  Consider optimizing via caching.
		// The sort property in use might not be the actual name of a field on this entity class.  It is possible
		// that the property maps to a @SortProperty annotation on one of the fields in this class.

		String embeddedSortProperty = null;

		for (Field field : FieldUtils.getDeclaredFields(entityClass)) {
			if (field.getName().equals(sortProperty)) {
				// Simple case--sort property maps directly to an entity field.
				break;
			}

			SortProperty sortAnnot = field.getAnnotation(SortProperty.class);
			if (sortAnnot != null && sortAnnot.value().equals(sortProperty)) {
				// The sort expression provided matches the sort annotation on this field.  Substitute field name.
				sortProperty = field.getName();
				break;
			}

			EmbeddedId embeddedId = field.getAnnotation(EmbeddedId.class);
			if (embeddedId != null) {
				Field[] embeddedFields = field.getType().getDeclaredFields();
				for (Field embeddedField : embeddedFields) {
					embeddedField.setAccessible(true);

					sortAnnot = embeddedField.getAnnotation(SortProperty.class);
					if (sortAnnot != null && sortAnnot.value().equals(sortProperty)) {
						// Found a sort property match on an embedded field.
						embeddedSortProperty = field.getName() + "." + embeddedField.getName();
						break;
					}
				}
			}
		}
		return embeddedSortProperty;
	}

	/**
	 * Applies default sort settings on an entity based on annotations.
	 *
	 * @param entityClass
	 * @param criteria
	 * @param <T>
	 */
	private <T extends IEntity> void applyDefaultSortToEntity(Class<T> entityClass, Criteria criteria) {
		List<SortSpecification> sorts = DaoUtils.getDefaultSortSpecifications(entityClass);
		if (sorts != null) {
			for (SortSpecification sort : sorts) {

				if (logger.isDebugEnabled()) {
					logger.debug(MessageFormat.format("Applying default sort: property={0}, ascending={1}",
							sort.getProperty(), sort.isAscending()));
				}

				if (sort.isAscending()) {
					criteria.addOrder(Order.asc(sort.getProperty()));
				} else {
					criteria.addOrder(Order.desc(sort.getProperty()));
				}
			}
		}
	}

	/**
	 * Will create a table alias in the criteria for the join table and
	 * replace the table part of the sortProperty with the newly created alias.
	 *
	 * @param criteria
	 * @param tableAliases
	 *@param sortProperty  @return
	 */
	private String getSortPropertyForJoinedFieldAndCreateTableAlias(Criteria criteria, List<String> tableAliases, String sortProperty) {
		//Deal with sorting by properties in joined tables, e.g Customer.orders.createdDate
		if (sortProperty.contains(".")) {
			String pathToAlias = StringUtils.substringBeforeLast(sortProperty, ".");
			String propertyName = StringUtils.substringAfterLast(sortProperty, ".");

			if(! tableAliases.contains(pathToAlias))
			{
				criteria.createAlias(pathToAlias, pathToAlias);
				tableAliases.add(pathToAlias);
			}

			sortProperty = pathToAlias + "." + propertyName;

		}
		return sortProperty;
	}

	public <T extends IEntity> Serializable insert(T entity) throws DuplicateDataException {
		try {
			getSession().save(entity);
		}
		catch (ConstraintViolationException cve) {
			//TODO: not all constraints are unique constraints
			throw new DuplicateDataException(cve.getMessage(), cve);
		}

		return entity.getId();
	}

	public <T extends IEntity> void saveOrUpdate(T entity) throws DuplicateDataException {
		try {
			getSession().saveOrUpdate(entity);
		}
		catch (ConstraintViolationException cve) {
			//TODO: not all constraints are unique constraints
			throw new DuplicateDataException(cve.getMessage(), cve);
		}
	}

	public <T extends IEntity> void update(T entity) throws DuplicateDataException {
		try {
			getSession().update(entity);
		}
		catch (ConstraintViolationException cve) {
			//TODO: not all constraints are unique constraints
			throw new DuplicateDataException(cve.getMessage(), cve);
		}
	}

	public <T extends IEntity> void delete(T entity) {
		getSession().delete(entity);
	}

	public <T extends IEntity> void deleteById(Class<T> entityClass, Serializable id) {
		StringBuilder deleteSql = new StringBuilder();
		deleteSql.append("delete from ");
		deleteSql.append(entityClass.getName());
		deleteSql.append(" where id = :id");

		Query delete = getSession().createQuery(deleteSql.toString());
		delete.setParameter("id", id);

		int count = delete.executeUpdate();

		if (count == 0) {
			throw new DataNotFoundException(entityClass, id);
		}
	}

	public <T extends IEntity> void deleteAll(Collection<T> entities) {
		if (entities.size() == 0)
			return;

		Class entityClass = null;

		Serializable[] entityIds = new Serializable[entities.size()];
		int i = 0;
		for (T entity : entities) {
			if (entityClass == null) {
				entityClass = entity.getClass();
			}
			entityIds[i++] = entity.getId();
		}

		StringBuilder deleteSql = new StringBuilder();
		deleteSql.append("delete from ");
		deleteSql.append(entityClass.getName());
		deleteSql.append(" where id in ( :ids )");

		Query delete = getSession().createQuery(deleteSql.toString());
		delete.setParameterList("ids", entityIds);

		int count = delete.executeUpdate();

		if (count != entities.size()) {
			throw new DataNotFoundException(entityClass, null);
		}
	}

	public Long getCounterValue(String counterName) {
		String schemaName = getDefaultSchemaName();
		if (!StringUtils.isBlank(schemaName)) {
			counterName = schemaName + "." + counterName;
		}

		BigDecimal dec = (BigDecimal) getSession().createSQLQuery("select " + counterName + ".nextval from dual").uniqueResult();
		return dec.longValue();
	}

	public void saveChanges() {
		getSession().flush();
	}

	public <T extends IEntity> List<T> executeQuery(String query, Object... params) {
		Query hql = getSession().createQuery(query);

		applyParamsToQuery(hql, params);

		return hql.list();
	}

	public List<Map<String, ?>> executeQueryProjections(String query, Object... params) {
		Query hql = getSession().createQuery(query);

		applyParamsToQuery(hql, params);

		return _executeProjectionsInternal(hql, params);
	}

	public <T> List<T> executeQueryProjection(String query, Object... params) {
		Query hql = getSession().createQuery(query);

		applyParamsToQuery(hql, params);

		List<Object[]> results = hql.list();

		List<T> projections = new ArrayList<T>();

		for (int i = 0; i < results.size(); i++) {

			Object obj = results.get(i);

			if (obj != null) {
				// If only one field projected, each result is the value projected.
				projections.add((T) obj);
			}
		}

		return projections;
	}

	private List<Map<String, ?>> _executeProjectionsInternal(Query hql, Object... params) {
		// Get the set of aliased fields in the projection.  These will be the keys in each map of name/value pairs.
		String[] aliases = hql.getReturnAliases();

		List results = hql.list();

		List<Map<String, ?>> mapResults = new ArrayList<Map<String, ?>>();

		for (int i = 0; i < results.size(); i++) {

			Object obj = results.get(i);

			Map<String, Object> map = new HashMap<String, Object>();

			if (obj != null) {
				if (aliases.length > 1) {
					// If multiple fields were projected in the query, each query result is an object array.
					Object[] arr = (Object[]) obj;
					for (int j = 0; j < aliases.length; j++) {
						map.put(aliases[j], arr[j]);
					}
				} else {
					// If only one field projected, each result is the value projected.
					map.put(aliases[0], obj);
				}
			}

			mapResults.add(map);
		}

		return mapResults;
	}

	public <T extends IEntity> List<T> executeNamedQuery(String name, Object params) {
		Query hql = getSession().getNamedQuery(name);

		applyParamsToQuery(hql, params);

		return hql.list();
	}

	public List<Map<String, ?>> executeNamedQueryProjections(String query, Object params) {
		Query hql = getSession().getNamedQuery(query);

		applyParamsToQuery(hql, params);

		return _executeProjectionsInternal(hql);
	}

	public <T> List<T> executeNamedQueryProjection(String query, Class returnType, Object... params) {
		Query hql = getSession().getNamedQuery(query);

		applyParamsToQuery(hql, params);

		List results = hql.list();

		List<T> projections = new ArrayList<T>();

		for (int i = 0; i < results.size(); i++) {

			Object obj = results.get(i);

			if (obj != null) {
				projections.add((T) obj);
			}
		}

		return projections;
	}

	public List<Map<String, ?>> executeSQL(String sql, Map<String, Class<?>> returnValues, Object... params) {
		SQLQuery query = getSession().createSQLQuery(sql);

		// Tell hibernate what the return values and their types are.
		for (String key : returnValues.keySet()) {
			Type type = getTypeForClass(returnValues.get(key));
			query.addScalar(key, type);
		}

		applyParamsToQuery(query, params);

		// Use a sorting map to determine the ordering of aliases in the source SQL's select statement.
		TreeMap<Integer, String> aliasMap = new TreeMap<Integer, String>();
		for (String alias : returnValues.keySet()) {
			int pos = sql.indexOf(alias);
			if (pos == -1) {
				throw new RuntimeException("Alias '" + alias + "' not found in provided SQL: " + sql);
			}

			aliasMap.put(pos, alias);
		}

		// This list of aliases will be used to convert the object array based return values into mapped values.
		ArrayList<String> orderedAliases = new ArrayList<String>(aliasMap.values());

		ArrayList<Map<String, ?>> results = new ArrayList<Map<String, ?>>();

		for (Object result : query.list()) {

			Object[] rawResult = (Object[]) result;

			HashMap<String, Object> mappedResults = new HashMap<String, Object>();

			int index = 0;
			for (String alias : orderedAliases) {
				mappedResults.put(alias, rawResult[index++]);
			}

			results.add(mappedResults);
		}

		return results;
	}

	public <T> List<T> executeSQLQuery(String sql, String returnAlias, Class returnType, Object... params) {
		SQLQuery query = getSession().createSQLQuery(sql);

		query.addScalar(returnAlias, getTypeForClass(returnType));

		applyParamsToQuery(query, params);

		ArrayList<T> results = new ArrayList<T>();

		for (Object result : query.list()) {
			results.add((T) result);
		}

		return results;
	}

	public int executeSQLUpdate(String sql, Object... params) {
		SQLQuery query = getSession().createSQLQuery(sql);

		applyParamsToQuery(query, params);

		return query.executeUpdate();
	}

	public <T extends IEntity> List<T> executeSQLQuery(Class<T> entityClass, String sql, Object... params) {
		SQLQuery query = getSession().createSQLQuery(sql);

		query.addEntity(entityClass);

		applyParamsToQuery(query, params);

		return query.list();
	}

	public List<Map<String, ?>> executeNamedSQL(String name, Object... params) {
		Query hql = getSession().getNamedQuery(name);

		applyParamsToQuery(hql, params);

		// Use the underlying hibernate session implementation to get access to the definition of the named query.
		// This will provide us with the list return value aliases that will be used to build a Map of each return
		// value.
		NamedSQLQueryDefinition queryDef = ((SessionFactoryImpl) sessionFactory).getNamedSQLQuery(name);
		NativeSQLQueryReturn[] returns = queryDef.getQueryReturns();

		// This list of aliases will be used to convert the object array based return values into mapped values.
		ArrayList<String> orderedAliases = new ArrayList<String>();
		for (NativeSQLQueryReturn queryReturn : returns) {
			if (!(queryReturn instanceof NativeSQLQueryScalarReturn)) {
				throw new RuntimeException("Only scalar return types supported. Why are you getting all fancy with your named SQL?");
			}
			orderedAliases.add(((NativeSQLQueryScalarReturn) queryReturn).getColumnAlias());
		}

		ArrayList<Map<String, ?>> results = new ArrayList<Map<String, ?>>();

		for (Object result : hql.list()) {
			Object[] rawResults = (Object[]) result;

			HashMap<String, Object> mappedResults = new HashMap<String, Object>();

			int index = 0;
			for (String alias : orderedAliases) {
				mappedResults.put(alias, rawResults[index++]);
			}

			results.add(mappedResults);
		}

		return results;
	}

	public <T extends IEntity> List<T> executeNamedEntitySQL(String name, Object... params) {
		Query hql = getSession().getNamedQuery(name);

		applyParamsToQuery(hql, params);

		return hql.list();
	}

	/**
	 * @param hql
	 * @param params
	 * @return
	 */
	public int executeUpdate(String hql, Object... params) {
		Query update = getSession().createQuery(hql);

		applyParamsToQuery(update, params);

		return update.executeUpdate();
	}
}
